-- $Id: sem-compunit-walkstatements-wf_delay_until.adb 15520 2010-01-07 12:53:45Z spark $
--------------------------------------------------------------------------------
-- (C) Altran Praxis Limited
--------------------------------------------------------------------------------
--
-- The SPARK toolset is free software; you can redistribute it and/or modify it
-- under terms of the GNU General Public License as published by the Free
-- Software Foundation; either version 3, or (at your option) any later
-- version. The SPARK toolset is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
-- Public License for more details. You should have received a copy of the GNU
-- General Public License distributed with the SPARK toolset; see file
-- COPYING3. If not, go to http://www.gnu.org/licenses for a complete copy of
-- the license.
--
--==============================================================================


separate (Sem.CompUnit.WalkStatements)

procedure wf_delay_until (Node  : in STree.SyntaxNode;
                          Scope : in Dictionary.Scopes)
is
   EncUnitSym   : Dictionary.Symbol;
   ResultType   : ExpRecord;
   RefVar       : SeqAlgebra.Seq;

begin
   -- ASSUME Node = delay_statement, nextderivative (derivative () = delay_expression

   SeqAlgebra.CreateSeq (TheHeap, RefVar);

   -- Process the expression of the delay_statement. Must be of type Ada.Real_Time.
   WalkExpression (ExpNode               => Next_Sibling (Child_Node (Node)),
                   Scope                 => Scope,
                   TypeContext           => Dictionary.GetPredefinedTimeType,
                   ContextRequiresStatic => False,
                   -- to get
                   Result                => ResultType,
                   RefVar                => RefVar,
                   ComponentData         => GlobalComponentData);

   -- Verify that ResultType is compatible with Ada.Real_Time
   if not Dictionary.CompatibleTypes (Scope => Scope,
                                      Left  => ResultType.TypeSymbol,
                                      Right => Dictionary.GetPredefinedTimeType) then
      ErrorHandler.SemanticError (866,
                                  ErrorHandler.NoReference,
                                  NodePosition (Next_Sibling (Child_Node (Node))),
                                  LexTokenManager.Null_String);
   end if;

   -- Verify that if the enclosing unit is a procedure then
   -- it has a delay annotation. Other enclosing units are invalid.
   -- Also mark the enclosing unit as having seen a delay statement.
   EncUnitSym := Dictionary.GetEnclosingCompilationUnit (Scope);

   if Dictionary.IsProcedure (EncUnitSym) and
      not Dictionary.IsOrIsInProtectedScope (Scope) then

      if Dictionary.HasDelayProperty (EncUnitSym) then
         Dictionary.MarkAccountsForDelay (EncUnitSym);

      else
         ErrorHandler.SemanticError (867,
                                     ErrorHandler.NoReference,
                                     NodePosition (Node),
                                     Dictionary.GetSimpleName (EncUnitSym));
      end if;

   elsif not Dictionary.IsTaskType (EncUnitSym) then
      -- Cannot call delay from the enclosing construct.
      ErrorHandler.SemanticError (907,
                                  ErrorHandler.NoReference,
                                  NodePosition (Node),
                                  LexTokenManager.Null_String);
   end if;

   -- NullVariable is an out stream and so need to be considered an import
   AddStreamEffects (Table,
                     TheHeap,
                     Node,
                     Dictionary.GetNullVariable,
                     RefVar);
   -- add reference variable list to RefList hash table for the statement node
   RefList.AddRelation (Table,
                        TheHeap,
                        Node,
                        Dictionary.GetNullVariable, -- ref vars dumped in data sink
                        RefVar);
end wf_delay_until;
